Attribute VB_Name = "modGraphics"

Option Explicit

' global direct3d object
Public D3D As Direct3D8
' the global direct3d helper object
Public D3DX As New D3DX8
' the rendering device
Public D3Ddevice As Direct3DDevice8

' the number of frames per second
Public FPS As Single

' used to measure time accurate to the millisecond
Public Declare Function timeGetTime Lib "winmm.dll" () As Long
' used to show/hide the windows cursor
Private Declare Function ShowCursor Lib "user32" (ByVal bShow As Long) As Long

' file paths used for loading resources
Public TexturePath As String
Public XfilePath As String

' error log
Public InitReport As String

' x file errors
Enum X_ERROR
    X_OK
    XERR_FILENOTFOUND
    XERR_SEEREPORT
    XERR_UNKNOWN
End Enum




' initializes direct3d and the rendering device, returning sucess/failure
Function Init(hWnd As Long, Device As ENGINEX_DEVICE, Optional SwapEffect As CONST_D3DSWAPEFFECT = D3DSWAPEFFECT_DISCARD) As Boolean
Dim D3Ddm As D3DDISPLAYMODE
Dim D3Dpp As D3DPRESENT_PARAMETERS
Dim DeviceType As CONST_D3DDEVTYPE
On Error GoTo InitFailed
    If Device.Fullscreen Then ShowCursor 0
    InitReport = "Loading graphics engine..."
    ' initialize arrays and file paths
    AddError "Initializing objects"
    ReDim Material(0)
    ReDim Texture(0)
    ReDim Camera(0)
    ReDim Light(0)
    WorldFrame.Tag = "Root Frame"
    AddError "Initializing file paths"
    TexturePath = App.Path & "\"
    XfilePath = App.Path & "\"
    ' get a pointer to direct3d
    AddError "Getting pointer to Direct3D"
    Set D3D = DX.Direct3DCreate()
    If D3D Is Nothing Then
        AddError "FATAL ERROR! Failed to get pointer to Direct3D! Graphics engine initialization aborted!"
        Exit Function
    End If
    ' set direct3d present parameters
    With D3Dpp
        AddError "Getting monitor information"
        D3D.GetAdapterDisplayMode D3DADAPTER_DEFAULT, D3Ddm
        AddError "Setting Direct3D presentation parameters"
        .BackBufferWidth = Device.DisplayMode.Width
        .BackBufferHeight = Device.DisplayMode.Height
        .BackbufferFormat = Device.DisplayMode.Format
        .BackBufferCount = 1
        .EnableAutoDepthStencil = 1
        .AutoDepthStencilFormat = D3DFMT_D16
        .SwapEffect = SwapEffect
        .Windowed = Not Device.Fullscreen
    End With
    If Device.Hardware Then DeviceType = D3DDEVTYPE_HAL Else DeviceType = D3DDEVTYPE_REF
    ' create a direct3d rendering device
    AddError "Getting pointer to Direct3D rendering device"
    Set D3Ddevice = D3D.CreateDevice(D3DADAPTER_DEFAULT, DeviceType, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, D3Dpp)
    If D3Ddevice Is Nothing Then
        AddError "FATAL ERROR! Failed to create rendering device! Graphics engine initialization aborted!"
        Exit Function
    End If
    AddError "Setting rendering options"
    D3Ddevice.SetRenderState D3DRS_CULLMODE, D3DCULL_NONE
    D3Ddevice.SetRenderState D3DRS_ZENABLE, D3DZB_TRUE
    D3Ddevice.SetRenderState D3DRS_FILLMODE, D3DFILL_SOLID
    D3Ddevice.SetRenderState D3DRS_LIGHTING, 1
    'D3Ddevice.SetRenderState D3DRS_AMBIENT, RGB(100, 100, 100)
    Init = Not Err.Number
    AddError "Initialization complete!"
    Exit Function
InitFailed:
    AddError "FATAL ERROR! An unknown error occured during graphics engine initialization! Aborting!"
End Function
' adds a string to init error log
Private Sub AddError(strError As String)
On Error Resume Next
    InitReport = InitReport & strError & vbNewLine
End Sub



' clears the backbuffer surface
Sub Clear(Optional Color As Long = vbBlack)
On Error Resume Next
    D3Ddevice.Clear 0, ByVal 0, D3DCLEAR_TARGET Or D3DCLEAR_ZBUFFER, Color, 1#, 0
End Sub



' renders all the frames in the engine
Sub Render()
Dim l As Long
On Error Resume Next
    Set lastMaterial = Nothing
    Set lastTexture = Nothing
    D3Ddevice.SetTexture 0, Nothing
    D3Ddevice.BeginScene
    modGeometry.SetWorldMatrix Vector3(0, 0, 0), Vector3(0, 0, 0), Vector3(1, 1, 1)
    RenderFrameAndChildren WorldFrame, 1
    D3Ddevice.EndScene
End Sub



' recursive function - renders a frame, then calls itself to render
' it's children until there are no more children, then it looks back at
' the stack to trace itself back up the tree until it can come down again.
' this continues until there are no more children in the original frame.
' if the frames are too many levels deep, the program may run out of stack
' space, but most worlds do not need more than 3 or 4 levels of frames
Sub RenderFrameAndChildren(Frame As cFrame, LocationIndex As Long)
Dim l As Long
Dim matTemp As D3DMATRIX
On Error Resume Next
    If LocationIndex > Frame.NumLocations Then Exit Sub
    D3Ddevice.GetTransform D3DTS_WORLD, matTemp
    ' set location as centre of the world
    Frame.Location(LocationIndex).SetAsCurrent
    ' render
    Frame.Render
    ' render children
    If Frame.NumChildren Then
        For l = 1 To Frame.NumChildren
            RenderFrameAndChildren Frame.ChildFrame(l), 1
        Next
    End If
    ' reset world matrix
    D3Ddevice.SetTransform D3DTS_WORLD, matTemp
    RenderFrameAndChildren Frame, LocationIndex + 1
End Sub




' flips the contents of the backbuffer onto the frontbuffer
Sub Present()
Static lastTime As Long
Dim thisTime As Long
On Error Resume Next
    D3Ddevice.Present ByVal 0, ByVal 0, 0, ByVal 0
    ' calculate FPS
    thisTime = timeGetTime
    FPS = 1000 / (thisTime - lastTime)
    lastTime = thisTime
End Sub



' cleans up direct3d, the frame hierarchy and the resources used
Sub CleanUp()
Dim l As Long
On Error Resume Next
    ShowCursor 1
    ' unload in reverse order to creation
    Set WorldFrame = Nothing ' this recursively sets all frames to nothing
    For l = LBound(Material) To UBound(Material)
        Set Material(l) = Nothing
    Next
    For l = LBound(Texture) To UBound(Texture)
        Set Texture(l) = Nothing
    Next
    For l = LBound(Light) To UBound(Light)
        Set Light(l) = Nothing
    Next
    For l = LBound(Camera) To UBound(Camera)
        Set Camera(l) = Nothing
    Next
    Set D3Ddevice = Nothing
    Set D3DX = Nothing
    Set DX = Nothing
End Sub
